/**************************************************************************
 * SPV Provisioning
 *
 * This program allows you to restore the registry after a buggy full reset
 * and will install a test certificate before restoring these settings 
 *************************************************************************/
  
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <windows.h>

#include "sha1.h"
#include "base64.h"

// ----------------------------------------------------------------------------------------------

#define CERT_FILE ".\\Cert\\SPVcert.cer"

#define FACTORY_PROVISIONING "\\Windows\\mxip_smartfon_1.provxml"
#define OPERATOR_PROVISIONING "\\Windows\\mxip_oem_10.provxml"

#define OTA_APP_CERTIFICATE_PROVISIONING \
	"<wap-provisioningdoc>\n\r" \
	"\t<characteristic type=\"CertificateStore\">\n\r" \
	"\t\t<characteristic type=\"Unprivileged Execution Trust Authorities\">\n\r" \
	"\t\t\t<characteristic type=\"%s\">\n\r" \
	"\t\t\t\t<parm name=\"EncodedCertificate\" value=\"%s\"/>\n\r" \
	"\t\t\t</characteristic>\n\r" \
	"\t\t</characteristic>\n\r" \
	"\t</characteristic>\n\r" \
	"</wap-provisioningdoc>" 

#define OTA_CAB_CERTIFICATE_PROVISIONING \
	"<wap-provisioningdoc>\n\r" \
	"\t<characteristic type=\"CertificateStore\">\n\r" \
	"\t\t<characteristic type=\"SPC\">\n\r" \
	"\t\t\t<characteristic type=\"%s\">\n\r" \
	"\t\t\t\t<parm name=\"EncodedCertificate\" value=\"%s\"/>\n\r" \
	"\t\t\t\t<parm name=\"Role\" value=\"16\"/>\n\r" \
	"\t\t\t</characteristic>\n\r" \
	"\t\t</characteristic>\n\r" \
	"\t</characteristic>\n\r" \
	"</wap-provisioningdoc>" 


#define RAPI_TIMEOUT 3000

// ----------------------------------------------------------------------------------------------

#define SPVPROV_NO_EXPORT 0
#define SPVPROV_NO_RAPI 1
#define SPVPROV_MAGIC_IMPORT 2
#define SPVPROV_OPEN_FILE 3
#define SPVPROV_GET_FILE_SIZE 4
#define SPVPROV_ALLOCATE 5
#define SPVPROV_READ 6
#define SPVPROV_CONNECTION 7
#define SPVPROV_AUTOMATION 8
#define SPVPROV_NOT_AUTHORIZED 9
#define SPVPROV_XML_INVALID 10
#define SPVPROV_BASE64_INVALID 11
#define SPVPROV_UNKNOWN 12
#define SPVPROV_CANCELLED 13
#define SPVPROV_SHA1_INIT 14
#define SPVPROV_SHA1_CALC 15
#define SPVPROV_SHA1_END 16
#define SPVPROV_BASE64_ENCODING 17
#define SPVPROV_NO_CERT 18
#define SPVPROV_PROV_STATUS 19

const char *ERRORS[] = {
	"Missing export %s !\n",
	"RAPI.dll not found \n",
	"CeRapiProvisioning export missing !\n",
	"Cannot open file %s\n",
	"Cannot retrieve file size\n",
	"Cannot allocate %d bytes\n",
	"Error reading file\n",
	"Error - check ActiveSync connection\n",
	"Automation error\n",
	"not authorized (is your phone unlocked ?)\n",
	"invalid XML file\n",
	"invalid Base 64 encoding\n",
	"Unknown status\n",
	"\nCancel\n",
	"Error initializing SHA-1\n",
	"Error computing SHA-1\n",
	"Error in last SHA-1 block\n",
	"Error during Base 64 encoding\n",	
	"User certificate not found\n",
	"Error, status %.4x - "
};

// ----------------------------------------------------------------------------------------------

#define SPVPROV_INFO_CNX 0
#define SPVPROV_INFO_CONTINUE 1
#define SPVPROV_INFO_GET_CERT 2
#define SPVPROV_INFO_BUILD_CERT_PROV 3
#define SPVPROV_INFO_APP_CERT_PROV 4
#define SPVPROV_INFO_USE_1 5
#define SPVPROV_INFO_USE_2 6
#define SPVPROV_INFO_FINISHED 7
#define SPVPROV_INFO_GET_SPV_PROV 8
#define SPVPROV_INFO_GET_ORANGE_PROV 9
#define SPVPROV_INFO_DO_SPV_PROV 10
#define SPVPROV_INFO_DO_ORANGE_PROV 11
#define SPVPROV_INFO_VERSION 12
#define SPVPROV_INFO_CAB_CERT_PROV 13


const char *INFOS[] = {
	"Connecting to Smartphone ... ",
	"Do you want to continue [Y/N] ",
	"Getting user certificate ... ",	 
	"Preparing certificate provisioning ... ",
	"Provisioning application certificate ... ",
	"Use : %s [(noProv)]\n",
	"If run with noProv, only the certificate provisioning will be done\n\n",
	"Finished !\n",
	"Getting phone provisioning file ... ",
	"Getting operator provisioning file ... ",	
	"Provisioning phone ... ",
	"Provisioning operator ... ",
	"SPVProv - Original SPV reprovisioning (US) [v1.1]",
	"Provisioning CABs certificate ... "
};	

// ----------------------------------------------------------------------------------------------

typedef struct _RAPIINIT { 
  DWORD cbSize; 
  HANDLE heRapiInit; 
  HRESULT hrRapiInit; 
} RAPIINIT; 

// ----------------------------------------------------------------------------------------------

typedef DWORD (WINAPI *tCeRapiInitEx)(RAPIINIT*);
tCeRapiInitEx CeRapiInitEx;

typedef DWORD (WINAPI *tCeRapiUninit)(void);
tCeRapiUninit CeRapiUninit;

typedef DWORD (WINAPI *tCeRapiProvisioning)(BSTR, DWORD, LPDWORD);
tCeRapiProvisioning CeRapiProvisioning;

typedef DWORD (WINAPI *tCeGetFileSize)(HANDLE, LPDWORD);
tCeGetFileSize CeGetFileSize;

typedef HANDLE (WINAPI *tCeCreateFile)(LPCWSTR, DWORD, DWORD, LPSECURITY_ATTRIBUTES, DWORD, DWORD, HANDLE);
tCeCreateFile CeCreateFile;

typedef DWORD (WINAPI *tCeReadFile)(HANDLE, LPVOID, DWORD, LPDWORD, LPOVERLAPPED);
tCeReadFile CeReadFile;

typedef BOOL (WINAPI *tCeCloseHandle)(HANDLE);
tCeCloseHandle CeCloseHandle;

// ----------------------------------------------------------------------------------------------

// Load imports from RAPI.DLL

#define IMPORT(Handle, Var, Type, Function) \
	Var = (Type)GetProcAddress(Handle, TEXT(Function)); \
	if (!Var) { \
		fprintf(stderr, ERRORS[SPVPROV_NO_EXPORT], Function); \
		return false; \
	}

BOOL init() {
	HMODULE rapiModule;

	rapiModule = LoadLibrary(TEXT("rapi.dll"));
	if (!rapiModule) {
		fprintf(stderr, ERRORS[SPVPROV_NO_RAPI]);
		return FALSE;
	}

	IMPORT(rapiModule, CeRapiInitEx, tCeRapiInitEx, "CeRapiInitEx");
	IMPORT(rapiModule, CeRapiUninit, tCeRapiUninit, "CeRapiUninit");
	IMPORT(rapiModule, CeGetFileSize, tCeGetFileSize, "CeGetFileSize");
	IMPORT(rapiModule, CeCreateFile, tCeCreateFile, "CeCreateFile");
	IMPORT(rapiModule, CeReadFile, tCeReadFile, "CeReadFile");	
	IMPORT(rapiModule, CeCloseHandle, tCeCloseHandle, "CeCloseHandle");

	/* <:)> */

	CeRapiProvisioning = (tCeRapiProvisioning)GetProcAddress(rapiModule, (LPCSTR)0x19);
	if (!CeRapiProvisioning) {
		fprintf(stderr, ERRORS[SPVPROV_MAGIC_IMPORT]);
		return FALSE;
	}

	/* </:)> */	

	return TRUE;
}

// ----------------------------------------------------------------------------------------------

// Download a file from the SPV

BOOL getRemoteFile(char *fileName, char **provText, DWORD *provSize) {
	HANDLE fileHandle;
	WCHAR  fileNameUnicode[MAX_PATH];
	DWORD sizeRead;

	MultiByteToWideChar(CP_ACP, 0, fileName, strlen(fileName) + 1, fileNameUnicode, sizeof(fileNameUnicode));

	fileHandle = CeCreateFile(fileNameUnicode, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if (fileHandle == INVALID_HANDLE_VALUE) {
		printf(ERRORS[SPVPROV_OPEN_FILE], fileName);
		return FALSE;
	}
	*provSize = CeGetFileSize(fileHandle, NULL);
	if (*provSize == 0xFFFFFFFF) {
		printf(ERRORS[SPVPROV_GET_FILE_SIZE]);
		CeCloseHandle(fileHandle);
		return FALSE;
	}
	*provText = (char*)calloc(*provSize + 1, sizeof(char));
	if (!*provText) {
		printf(ERRORS[SPVPROV_ALLOCATE], *provSize);
		CeCloseHandle(fileHandle);
		return FALSE;
	}
	if (!CeReadFile(fileHandle, (LPVOID)*provText, *provSize, &sizeRead, NULL) ||
	    sizeRead != *provSize) {
		printf(ERRORS[SPVPROV_READ]);
		CeCloseHandle(fileHandle);
		return FALSE;
	}
	CeCloseHandle(fileHandle);
	
	printf("OK\n");

	return TRUE;
}

// ----------------------------------------------------------------------------------------------

// Get a local file
		
BOOL getFile(char *fileName, char **fileContent, DWORD *fileSize) {
	FILE *provFile;

	provFile = fopen(fileName, "rb");
	if (!provFile) {
		printf(ERRORS[SPVPROV_OPEN_FILE], fileName);
		return FALSE;
	}
	fseek(provFile, 0, SEEK_END);
	*fileSize = ftell(provFile);
	fseek(provFile, 0, SEEK_SET);
	*fileContent = (char*)malloc(*fileSize);
	if (!*fileContent) {
		printf(ERRORS[SPVPROV_ALLOCATE], *fileSize);
		fclose(provFile);
		return FALSE;
	}
	fread(*fileContent, 1, *fileSize, provFile);
	fclose(provFile);

	printf("OK\n");

	return TRUE;
}

// ----------------------------------------------------------------------------------------------

// Open connection to the SPV

BOOL openConnection() {
	RAPIINIT rapiInit;

	rapiInit.cbSize = sizeof(RAPIINIT);
	printf(INFOS[SPVPROV_INFO_CNX]);
	CeRapiInitEx(&rapiInit);
	if (rapiInit.hrRapiInit != 0 || 
		WaitForSingleObject(rapiInit.heRapiInit, RAPI_TIMEOUT) != WAIT_OBJECT_0) {
			printf(ERRORS[SPVPROV_CONNECTION]);
			return FALSE;
	}
	printf("OK\n");

	return TRUE;
}

// ----------------------------------------------------------------------------------------------

// Send a provisioning WAP file to the SPV

BOOL sendProvisioning(char *text) {
	WCHAR *provUnicode;
	BSTR provString;
	DWORD provSize;
	DWORD provStatus;

	provUnicode = (WCHAR*)calloc(strlen(text) + 1, sizeof(unsigned short));
	if (!provUnicode) {
		printf(ERRORS[SPVPROV_ALLOCATE], (strlen(text) + 1) * sizeof(unsigned short));
		return FALSE;
	}
	provSize = MultiByteToWideChar(CP_ACP, 0, text, strlen(text) + 1, provUnicode, (strlen(text) + 1) * sizeof(unsigned short));
	provString = SysAllocString(provUnicode);
	if (!provString) {
		printf(ERRORS[SPVPROV_AUTOMATION]);
		return FALSE;
	}
	free(provUnicode);

	provStatus = CeRapiProvisioning(provString, 1, &provSize);

	SysFreeString(provString);

	if (provStatus == 0) 
		printf("OK\n");
	else {
		printf(ERRORS[SPVPROV_PROV_STATUS], provStatus);
		switch(provStatus) {
			case 0x80070005:
				printf(ERRORS[SPVPROV_NOT_AUTHORIZED]);
				break;
			case 0x80042009:
				printf(ERRORS[SPVPROV_XML_INVALID]);
				break;
			case 0x80004005:
				printf(ERRORS[SPVPROV_BASE64_INVALID]);
				break;
			default:
				printf(ERRORS[SPVPROV_UNKNOWN]);
		}
		return FALSE;
	}

	return TRUE;
}

// ----------------------------------------------------------------------------------------------

// Continue, Yes/No ?

BOOL checkContinue() {
	char x;

	printf(INFOS[SPVPROV_INFO_CONTINUE]);
	fflush(stdin);
	x = toupper(getchar());
	if (x != 'O') {
		printf(ERRORS[SPVPROV_CANCELLED]);
		return FALSE;
	}
	printf("\n");

	return TRUE;
}

// ----------------------------------------------------------------------------------------------

// Build a certificate OTA file

BOOL processCertificate() {
	char *certificate;
	DWORD certificateSize;
	char certificateCommand[4096];
	char base64Certificate[2000];
	unsigned int base64CertificateLength;
	SHA1Context SHA1State;
	unsigned char SHA1[SHA1HashSize];
	char SHA1Ascii[2 * SHA1HashSize + 1];
	int i;

	/* Get user certificate */

	printf(INFOS[SPVPROV_INFO_GET_CERT]);
	if (!getFile(CERT_FILE, &certificate, &certificateSize))
		return FALSE;

	/* Get SHA1 hash of certificate */

	printf(INFOS[SPVPROV_INFO_BUILD_CERT_PROV]);
	if (SHA1Reset(&SHA1State) != shaSuccess) {
		printf(ERRORS[SPVPROV_SHA1_INIT]);
		return FALSE;
	}
	if (SHA1Input(&SHA1State, (const uint8_t*)certificate, certificateSize) != shaSuccess) {
		printf(ERRORS[SPVPROV_SHA1_CALC]);
		return FALSE;
	}
	if (SHA1Result(&SHA1State, (uint8_t*)SHA1) != shaSuccess) {
		printf(ERRORS[SPVPROV_SHA1_END]);
		return FALSE;
	}

	for (i=0; i<SHA1HashSize; i++) 
		sprintf(&SHA1Ascii[2 * i], "%.2X", SHA1[i]);

	printf("OK\n");

	/* Get base 64 encoded certificate */

	memset(base64Certificate, 0, sizeof(base64Certificate));
	if (encode64(certificate, certificateSize, base64Certificate,
			sizeof(base64Certificate), &base64CertificateLength) != OK) {
		printf(ERRORS[SPVPROV_BASE64_ENCODING]);
		return FALSE;
	}

	sprintf(certificateCommand, OTA_APP_CERTIFICATE_PROVISIONING, SHA1Ascii, base64Certificate);

	printf(INFOS[SPVPROV_INFO_APP_CERT_PROV]);

	if (!sendProvisioning(certificateCommand))
		return FALSE;

	sprintf(certificateCommand, OTA_CAB_CERTIFICATE_PROVISIONING, SHA1Ascii, base64Certificate);

	printf(INFOS[SPVPROV_INFO_CAB_CERT_PROV]);

	if (!sendProvisioning(certificateCommand))
		return FALSE;

	return TRUE;
}


// ----------------------------------------------------------------------------------------------

int main(int argc, char* argv[])
{
	FILE *checkCertificate;
	char *deviceProvisioning;
	DWORD deviceProvisioningSize;
	char *operatorProvisioning;
	DWORD operatorProvisioningSize;

	printf("%s\n\n", INFOS[SPVPROV_INFO_VERSION]);

	// Check args
	if (argc >= 2 && strcmp(argv[1], "noProv") != 0) {
		printf(INFOS[SPVPROV_INFO_USE_1], argv[0]);
		printf(INFOS[SPVPROV_INFO_USE_2]);		
		return 1;
	}

	// Initialize 
	if (!init()) 
		return 1;
	
	// Open connection
	if (!openConnection()) {
		CeRapiUninit();
		return 1;
	}

	// Check certificate
	checkCertificate = fopen(CERT_FILE, "rb");
	if (!checkCertificate) {
		printf(ERRORS[SPVPROV_NO_CERT]);
		if (!checkContinue()) {
			CeRapiUninit();
			return 1;
		}
	}
	else {
		fclose(checkCertificate);

		// Certificate provisioning

		if (!processCertificate()) {
			CeRapiUninit();
			return 1;	
		}
	}

	if (argc == 2 && strcmp(argv[1], "noProv") == 0) {
		printf(INFOS[SPVPROV_INFO_FINISHED]);

		return 0;
	}
		

	// Base provisioning

	printf(INFOS[SPVPROV_INFO_GET_SPV_PROV]);
	if (!getRemoteFile(FACTORY_PROVISIONING, &deviceProvisioning,
				&deviceProvisioningSize)) {
		CeRapiUninit();
		return 1;
	}
	printf(INFOS[SPVPROV_INFO_GET_ORANGE_PROV]);
	if (!getRemoteFile(OPERATOR_PROVISIONING, &operatorProvisioning,
				&operatorProvisioningSize)) {
		CeRapiUninit();
		return 1;
	}

	printf(INFOS[SPVPROV_INFO_DO_SPV_PROV]);
	if (!sendProvisioning(deviceProvisioning)) {
		CeRapiUninit();
		return 1;
	}

	printf(INFOS[SPVPROV_INFO_DO_ORANGE_PROV]);
	if (!sendProvisioning(operatorProvisioning)) {
		CeRapiUninit();
		return 1;
	}

	printf(INFOS[SPVPROV_INFO_FINISHED]);

	return 0;
}


